@@ -3,7 +3,9 @@
 import ast
 import base64
 import hashlib
+import itertools
 import os
+import pathlib
 import pprint
 import shutil
 import tempfile
@@ -60,25 +62,34 @@ class WheelRepair:
                  extract_dir: typing.Optional[str] = None,
                  add_dlls: typing.Optional[typing.Set[str]] = None,
                  no_dlls: typing.Optional[typing.Set[str]] = None,
+                 ignore_in_wheel: bool = False,
                  verbose: int = 0) -> None:
         """Initialize a wheel repair object.
         whl_path: Path to the wheel to repair
-        extract_dir: Directory where intermediate files are created in the
-            process of repairing the wheel. If None, a temporary directory is
-            created.
+        extract_dir: Directory where wheel is extracted. If None, a temporary
+            directory is created.
         dest_dir: Directory to place the repaired wheel. If None, it defaults to
             wheelhouse, relative to the current working directory.
         add_dlls: Set of lowercase DLL names to force inclusion into the wheel
         no_dlls: Set of lowercase DLL names to force exclusion from wheel
             (cannot overlap with add_dlls)
         no_mangles: Set of lowercase DLL names not to mangle
+        ignore_in_wheel: whether to ignore DLLs that are already in the wheel
         verbose: verbosity level, 0 to 3"""
         if not os.path.isfile(whl_path):
             raise FileNotFoundError(f'{whl_path} not found')
+
+        self._verbose = verbose
         self._whl_path = whl_path
         self._whl_name = os.path.basename(whl_path)
-        self._distribution_name = self._whl_name[:self._whl_name.index('-')]
-        self._version = self._whl_name.split('-')[1]
+        if not self._whl_name.endswith('.whl'):
+            raise ValueError(f'{self._whl_name} is not a valid wheel name')
+        whl_name_split = os.path.splitext(self._whl_name)[0].split('-')
+        if len(whl_name_split) not in (5, 6):
+            raise ValueError(f'{self._whl_name} is not a valid wheel name')
+        self._distribution_name = whl_name_split[0]
+        self._version = whl_name_split[1]
+
         if extract_dir is None:
             # need to assign temp directory object to an attribute to prevent it
             # from being destructed
@@ -87,9 +98,44 @@ class WheelRepair:
         else:
             self._extract_dir_obj = None
             self._extract_dir = extract_dir
+        try:
+            shutil.rmtree(self._extract_dir)
+        except FileNotFoundError:
+            pass
+        os.makedirs(self._extract_dir)
+        if self._verbose >= 1:
+            print(f'extracting {self._whl_name} to {self._extract_dir}')
+        with zipfile.ZipFile(self._whl_path) as whl_file:
+            whl_file.extractall(self._extract_dir)
+
         self._add_dlls = set() if add_dlls is None else add_dlls
         self._no_dlls = set() if no_dlls is None else no_dlls
-        self._verbose = verbose
+
+        # Modify self._no_dlls to include those that are already part of every
+        # Python distribution the wheel targets.
+        abi_tags = whl_name_split[-2].split('.')
+        platform_tags = whl_name_split[-1].split('.')
+        ignore_by_distribution = set().union(*dll_list.ignore_by_distribution.values())
+        for abi_platform in itertools.product(abi_tags, platform_tags):
+            abi_platform = '-'.join(abi_platform)
+            if abi_platform in dll_list.ignore_by_distribution:
+                ignore_by_distribution &= dll_list.ignore_by_distribution[abi_platform]
+            else:
+                ignore_by_distribution = set()
+                break
+        self._no_dlls |= ignore_by_distribution
+
+        # If ignore_in_wheel is True, save list of all directories in the wheel.
+        # These directories will be used to search for DLLs that are already in
+        # the wheel.
+        if ignore_in_wheel:
+            self._wheel_dirs = [self._extract_dir]
+            for root, dirnames, _ in os.walk(self._extract_dir):
+                for dirname in dirnames:
+                    self._wheel_dirs.append(os.path.join(root, dirname))
+        else:
+            self._wheel_dirs = None
+        self._ignore_in_wheel = ignore_in_wheel
 
     @staticmethod
     def _rehash(file_path: str) -> typing.Tuple[str, int]:
@@ -190,26 +236,49 @@ class WheelRepair:
 
         Precondition: path must end with .pyd"""
         top_level_dirs = [
-            self._extract_dir,
-            os.path.join(self._extract_dir, f'{self._distribution_name}-{self._version}.data', 'purelib'),
-            os.path.join(self._extract_dir, f'{self._distribution_name}-{self._version}.data', 'platlib'),
+            pathlib.Path(self._extract_dir),
+            pathlib.Path(self._extract_dir) / f'{self._distribution_name}-{self._version}.data' / 'purelib',
+            pathlib.Path(self._extract_dir) / f'{self._distribution_name}-{self._version}.data' / 'platlib',
         ]
-        return any(os.path.samefile(os.path.dirname(path), p) for p in top_level_dirs if os.path.isdir(p))
+        return pathlib.Path(os.path.dirname(path)) in top_level_dirs
+
+    def _get_repair_version(self) -> str:
+        """If this wheel has already been repaired, return the delvewheel
+        version that performed the repair. Otherwise, return the empty
+        string."""
+        filename = os.path.join(self._extract_dir, f'{self._distribution_name}-{self._version}.dist-info', 'DELVEWHEEL')
+        if os.path.isfile(filename):
+            with open(filename) as file:
+                return file.read().strip()
+        return ''
+
+    def _split_dependency_paths(self, dependency_paths: typing.Iterable):
+        """Given an iterable of DLL paths, partition the contents into a tuple
+        of sorted lists
+        (dependency_paths_in_wheel, dependency_paths_outside_wheel).
+        dependency_paths_in_wheel contains the paths to DLLs that are already in
+        the wheel, and dependency_paths_outside_wheel contains the paths to DLLs
+        that are not in the wheel."""
+        dependency_paths_in_wheel = []
+        dependency_paths_outside_wheel = []
+        for dependency_path in dependency_paths:
+            if pathlib.Path(self._extract_dir) in pathlib.Path(dependency_path).parents:
+                dependency_paths_in_wheel.append(dependency_path)
+            else:
+                dependency_paths_outside_wheel.append(dependency_path)
+        dependency_paths_in_wheel.sort()
+        dependency_paths_outside_wheel.sort()
+        return dependency_paths_in_wheel, dependency_paths_outside_wheel
 
     def show(self) -> None:
         """Show the dependencies that the wheel has."""
         print(f'Analyzing {self._whl_name}\n')
 
-        # extract wheel
-        try:
-            shutil.rmtree(self._extract_dir)
-        except FileNotFoundError:
-            pass
-        os.makedirs(self._extract_dir)
-        if self._verbose >= 1:
-            print(f'extracting {self._whl_name} to {self._extract_dir}')
-        with zipfile.ZipFile(self._whl_path) as whl_file:
-            whl_file.extractall(self._extract_dir)
+        # check whether wheel has already been repaired
+        repair_version = self._get_repair_version()
+        if repair_version:
+            print(f'Delvewheel {repair_version} has already repaired this wheel.')
+            return
 
         # find dependencies
         dependency_paths = set()
@@ -221,25 +290,37 @@ class WheelRepair:
                 if filename.endswith('.pyd'):
                     extension_module_path = os.path.join(root, filename)
                     extension_module_paths.append(extension_module_path)
-                    discovered, ignored, not_found = patch_dll.get_all_needed(extension_module_path, self._add_dlls, self._no_dlls, 'ignore')
+                    discovered, ignored, not_found = patch_dll.get_all_needed(extension_module_path, self._add_dlls, self._no_dlls, self._wheel_dirs, 'ignore')
                     dependency_paths |= discovered
                     ignored_dll_names |= ignored
                     not_found_dll_names |= not_found
 
-        dependency_paths = list(dependency_paths)
-        dependency_paths.sort()
-        print('The following dependent DLLs will be copied into the wheel.')
-        if dependency_paths:
-            for dependency_path in dependency_paths:
+        if self._ignore_in_wheel:
+            dependency_paths_in_wheel, dependency_paths_outside_wheel = self._split_dependency_paths(dependency_paths)
+        else:
+            dependency_paths_in_wheel = None
+            dependency_paths_outside_wheel = list(dependency_paths)
+            dependency_paths_outside_wheel.sort()
+        print('The following DLLs will be copied into the wheel.')
+        if dependency_paths_outside_wheel:
+            for dependency_path in dependency_paths_outside_wheel:
                 print(f'    {os.path.basename(dependency_path)} ({dependency_path})')
             for not_found_dll_name in not_found_dll_names:
                 print(f'    {not_found_dll_name} (Error: Not Found)')
         else:
             print('    None')
 
+        if self._ignore_in_wheel:
+            print('\nThe following DLLs are already in the wheel and will not be copied.')
+            if dependency_paths_in_wheel:
+                for dependency_path in dependency_paths_in_wheel:
+                    print(f'    {os.path.basename(dependency_path)} ({os.path.join(self._whl_name, os.path.relpath(dependency_path, self._extract_dir))})')
+            else:
+                print('    None')
+
         ignored_dll_names = list(ignored_dll_names)
         ignored_dll_names.sort()
-        print('\nThe following dependent DLLs will not be copied into the wheel.')
+        print("\nThe following DLLs are assumed to be present in the end user's environment and will not be copied into the wheel.")
         if ignored_dll_names:
             for ignored_dll_name in ignored_dll_names:
                 print(f'    {ignored_dll_name}')
@@ -256,16 +337,11 @@ class WheelRepair:
         lib_sdir is the suffix for the directory to store the DLLs"""
         print(f'repairing {self._whl_path}')
 
-        # extract wheel
-        try:
-            shutil.rmtree(self._extract_dir)
-        except FileNotFoundError:
-            pass
-        os.makedirs(self._extract_dir)
-        if self._verbose >= 1:
-            print(f'extracting {self._whl_name} to {self._extract_dir}')
-        with zipfile.ZipFile(self._whl_path) as whl_file:
-            whl_file.extractall(self._extract_dir)
+        # check whether wheel has already been repaired
+        repair_version = self._get_repair_version()
+        if repair_version:
+            print(f'Delvewheel {repair_version} has already repaired this wheel.')
+            return
 
         # find dependencies and copy them into wheel
         print('finding DLL dependencies')
@@ -284,7 +360,7 @@ class WheelRepair:
                     elif self._verbose >= 1:
                         print(f'analyzing package-level extension module {os.path.relpath(extension_module_path, self._extract_dir)}')
                     extension_module_paths.append(extension_module_path)
-                    discovered, ignored = patch_dll.get_all_needed(extension_module_path, self._add_dlls, self._no_dlls)[:2]
+                    discovered, ignored = patch_dll.get_all_needed(extension_module_path, self._add_dlls, self._no_dlls, self._wheel_dirs)[:2]
                     dependency_paths |= discovered
                     ignored_dll_names |= ignored
         if not dependency_paths:
@@ -399,6 +475,12 @@ class WheelRepair:
                     if os.path.isdir(os.path.join(extra_dir, item)):
                         self._patch_init(os.path.join(extra_dir, item, '__init__.py'), libs_dir_name, load_order_filename)
 
+        # create .dist-info/DELVEWHEEL file to indicate that the wheel has been
+        # repaired
+        filename = os.path.join(self._extract_dir, f'{self._distribution_name}-{self._version}.dist-info', 'DELVEWHEEL')
+        with open(filename, 'w') as file:
+            file.write(f'{version.__version__}\n')
+
         # update record file, which tracks wheel contents and their checksums
         dist_info_foldername = '-'.join(self._whl_name.split('-')[:2]) + '.dist-info'
         record_filepath = os.path.join(self._extract_dir, dist_info_foldername, 'RECORD')
